
const persist = require('node-persist');
const {VestelNetworkRemote, VestelFollowTV} = require('vestel-network-remote');
const UPnPClient = require('@langovoi/upnp-device-client');

let AppleTV;

const key_map = {};

module.exports = function (homebridge) {
    const storage = persist.create();

    storage.initSync({
        dir: homebridge.user.persistPath(),
        stringify: data => JSON.stringify(data, undefined, 4),
    });

    try {
        AppleTV = require('node-appletv');
    } catch (err) {
        console.error('Failed to load node-appletv', err);
        console.log('node-appletv is not installed. You will not be able to use an Apple TV to turn a TV on.');
    }

    homebridge.registerAccessory('vestel-network-remote', 'TV', TVAccessory.use(homebridge, storage));

    Object.assign(key_map, {
        [homebridge.hap.Characteristic.RemoteKey.REWIND]: 'REWIND',
        [homebridge.hap.Characteristic.RemoteKey.FAST_FORWARD]: 'FAST_FORWARD',
        // [homebridge.hap.Characteristic.RemoteKey.NEXT_TRACK]:
        // [homebridge.hap.Characteristic.RemoteKey.PREVIOUS_TRACK]:
        [homebridge.hap.Characteristic.RemoteKey.ARROW_UP]: 'UP',
        [homebridge.hap.Characteristic.RemoteKey.ARROW_DOWN]: 'DOWN',
        [homebridge.hap.Characteristic.RemoteKey.ARROW_LEFT]: 'LEFT',
        [homebridge.hap.Characteristic.RemoteKey.ARROW_RIGHT]: 'RIGHT',
        [homebridge.hap.Characteristic.RemoteKey.SELECT]: 'SELECT',
        [homebridge.hap.Characteristic.RemoteKey.BACK]: 'BACK',
        [homebridge.hap.Characteristic.RemoteKey.EXIT]: 'EXIT',
        [homebridge.hap.Characteristic.RemoteKey.PLAY_PAUSE]: 'PLAY', // ???
        [homebridge.hap.Characteristic.RemoteKey.INFORMATION]: 'INFO',
    });
};

class TVAccessory {
    static use(homebridge, storage) {
        return class extends TVAccessory {
            static get homebridge_api() {
                return homebridge;
            }

            static get storage() {
                return storage;
            }
        };
    }

    static get hap() {
        return this.homebridge_api.hap;
    }

    constructor(log, config) {
        this.log = log;
        this.name = config.name;
        this.mac_address = config.mac_address;
        this.ip_address = config.ip_address;
        this.media_renderer_location = config.media_renderer_location !== undefined ?
            config.media_renderer_location : 'http://' + this.ip_address + ':2869/dmr.xml';

        this.storage_key = 'vestel-network-remote.TV.' + this.name + '.json';

        // Services
        this.expose_television_service = config.expose_television_service !== undefined ?
            config.expose_television_service : !!config.tv_sources;
        this.expose_television_speaker_service = config.expose_television_speaker_service !== undefined ?
            config.expose_television_speaker_service : this.expose_television_service && this.media_renderer_location;

        this.expose_power_service = config.expose_power_service !== undefined ? config.expose_power_service :
            !this.expose_television_service;
        this.expose_speaker_service = config.expose_speaker_service !== undefined ? config.expose_speaker_service :
            !this.expose_television_speaker_service && this.media_renderer_location;
        this.expose_speaker_lightbulb_service = config.expose_speaker_lightbulb_service !== undefined ?
            config.expose_speaker_lightbulb_service : false;

        if (!this.expose_power_service && !this.expose_speaker_service && !this.expose_speaker_lightbulb_service && !this.expose_television_service && !this.expose_television_speaker_service) {
            throw new Error('Trying to create a vestel-network-remote.TV accessory with no services');
        }

        // Connect to an Apple TV
        // This is used to turn the TV on
        if (config.apple_tv && AppleTV) {
            const credentials = typeof config.apple_tv === 'string'
                ? AppleTV.parseCredentials(config.apple_tv) : typeof config.apple_tv === 'array'
                ? AppleTV.parseCredentials(config.apple_tv.join(':')) : new AppleTV.Credentials(
                    config.apple_tv.unique_identifier,
                    Buffer.from(config.apple_tv.identifier, 'hex'),
                    Buffer.from(config.apple_tv.pairing_id, 'hex').toString(),
                    Buffer.from(config.apple_tv.public_key, 'hex'),
                    Buffer.from(config.apple_tv.encryption_key, 'hex')
                );

            AppleTV.scan(credentials.uniqueIdentifier)
                .then(devices => {
                    this.apple_tv = devices[0];
                    this.apple_tv.on('error', err => {
                        console.error(this.name + ' Apple TV error', err);
                    });
                    return this.apple_tv.openConnection(credentials);
                })
                .then(device => {
                    this.log('Connected to ' + config.name);
                })
                .catch(error => this.log(error));
        }

        // Connect to the TV
        this.remote = new VestelNetworkRemote(this.ip_address, config.remote_port);
        this.followtv = new VestelFollowTV(this.ip_address, config.status_port);

        this.handleMediaRendererEvent = this.handleMediaRendererEvent.bind(this);

        if (this.media_renderer_location) {
            this.media_renderer_client = new UPnPClient(this.media_renderer_location);
            this.media_renderer_client.on('error', err => console.error('[UPnP] Error', err));

            this.media_renderer_client.subscribe('RenderingControl', this.handleMediaRendererEvent);
        }

        // Add Input Source services
        this.tv_sources = [];

        for (let [source_id, data] of Object.entries(config.tv_sources || {})) {
            this.tv_sources.push(Object.assign(data instanceof Array ? data = {
                name: data[0],
                input_source_type: data[1],
                input_device_type: data[2],
            } : data, {source_id}));

            this.addInputSourceService(this.tv_sources.length - 1, data);
        }

        for (let [id, app_config] of Object.entries(config.tv_apps || {})) {
            this.addAppInputSourceService(id, app_config);
        }

        const tv_channel_input_source_services = [];

        for (let channel_number of config.tv_channels || []) {
            const name = (this.getItemSync('CachedChannelNames') || {})[channel_number] || 'Channel ' + channel_number;

            tv_channel_input_source_services.push(this.addAppInputSourceService(name, {
                name,
                input_source_type: this.constructor.hap.Characteristic.InputSourceType.TUNER,
                channel_number,
            }));
        }

        if (tv_channel_input_source_services.length) {
            this.followtv.getActiveChannelList().then(channels => {
                const cached_channel_names = {};

                for (let input_service of tv_channel_input_source_services) {
                    const channel = channels.find(channel => channel.rsn == input_service.config.channel_number);

                    if (!channel) continue;

                    input_service.setCharacteristic(this.constructor.hap.Characteristic.Name, channel.name);
                    cached_channel_names[input_service.config.channel_number] = channel.name;
                }

                return this.setItem('CachedChannelNames', cached_channel_names);
            }, err => this.log('Error fetching channel names', err));
        }

        if (this.expose_television_service) {
            this.tv_service.getCharacteristic(this.constructor.hap.Characteristic.ActiveIdentifier).setProps({
                maxValue: this.input_source_services.length,
            });
        }

        this.apple_tv_input = config.apple_tv_input || this.input_source_services.findIndex(service => service.getCharacteristic(this.constructor.hap.Characteristic.InputSourceType) === this.constructor.hap.Characteristic.InputSourceType.HDMI) || 1;
        this.default_input = config.default_input || 0;

        this.accessory_information_service = new this.constructor.hap.Service.AccessoryInformation();
        this.accessory_information_service
            .setCharacteristic(this.constructor.hap.Characteristic.Manufacturer, config.manufacturer || 'Samuel Elliott')
            .setCharacteristic(this.constructor.hap.Characteristic.Model, config.model || 'https://gitlab.fancy.org.uk/samuel/homebridge-vestel-network-remote')
            .setCharacteristic(this.constructor.hap.Characteristic.SerialNumber, config.serial_number || this.ip_address)
            .setCharacteristic(this.constructor.hap.Characteristic.FirmwareRevision, require('../package').version);

        if (this.expose_power_service || this.expose_television_service) {
            const updatePowerState = async () => {
                const on = await this.getPowerState();

                if (this.expose_power_service) this.power_service.getCharacteristic(this.constructor.hap.Characteristic.On)
                    .updateValue(on);
                if (this.expose_television_service) this.tv_service.getCharacteristic(this.constructor.hap.Characteristic.Active)
                    .updateValue(on ? this.constructor.hap.Characteristic.Active.ACTIVE
                        : this.constructor.hap.Characteristic.Active.INACTIVE);
            };

            updatePowerState();
            this.update_power_state_interval = setInterval(updatePowerState, 60000);
        }
    }

    async identify(callback) {
        this.log('Identify called');
        await this.remote.send('INFO');
        callback();
    }

    getServices() {
        const services = [this.accessory_information_service];

        if (this.expose_power_service) services.push(this.power_service);
        if (this.expose_speaker_service) services.push(this.speaker_service);
        if (this.expose_speaker_lightbulb_service) services.push(this.speaker_lightbulb_service);
        if (this.expose_television_service) services.push(this.tv_service, ...this.input_source_services);
        if (this.expose_television_speaker_service) services.push(this.tv_speaker_service);

        return services;
    }

    //
    // Storage
    //

    async getItem(key) {
        if (!this.storage_cache) {
            this.storage_cache = await this.constructor.storage.getItem(this.storage_key) || {};
        }

        return this.storage_cache[key];
    }

    getItemSync(key) {
        if (!this.storage_cache) {
            this.storage_cache = this.constructor.storage.getItemSync(this.storage_key) || {};
        }

        return this.storage_cache[key];
    }

    async setItem(key, value) {
        if (!this.storage_cache) {
            this.storage_cache = await this.constructor.storage.getItem(this.storage_key) || {};
        }

        this.storage_cache[key] = value;

        await this.constructor.storage.setItem(this.storage_key, this.storage_cache);
    }

    //
    // Power service
    //

    get power_service() {
        const power_service = new this.constructor.hap.Service.Switch(this.name);

        power_service.getCharacteristic(this.constructor.hap.Characteristic.On)
            .on('get', callback => this.getPowerState().then(data => callback(undefined, data)).catch(callback))
            .on('set', (on, callback) => this.setPowerState(on).then(data => callback(undefined, data)).catch(callback));

        return Object.defineProperty(this, 'power_service', {configurable: true, value: power_service}).power_service;
    }

    async getPowerState() {
        this.log('Getting power state of ' + this.name);

        // Try connecting to the TV
        // If it doesn't respond in 2 seconds assume the TV is off
        try {
            await Promise.race([
                this.remote.connect(async () => {}),
                new Promise((rs, reject) => setTimeout(reject, 2000)),
            ]);

            this.log(this.name + ' is on');

            return true;
        } catch (err) {
            this.log(this.name + ' is off');

            return false;
        }
    }

    setPowerState(on) {
        return this['setPowerStateO' + (on ? 'n' : 'ff')]();
    }

    async setPowerStateOn() {
        if (this.apple_tv) {
            this.log('Turning ' + this.name + ' on using Apple TV');

            try {
                await this.apple_tv.sendKeyCommand(AppleTV.AppleTV.Key.Menu);

                this.last_input_switched = this.apple_tv_input;
                if (this.expose_television_service) this.tv_service.getCharacteristic(this.constructor.hap.Characteristic.ActiveIdentifier).updateValue(this.apple_tv_input);
            } catch (err) {
                console.error(err);
                throw err;
            }

            return;
        }

        // Cannot turn the TV on without IR or HDMI CEC
        // Turning the TV off disconnects the network and WOL doesn't work
        throw new Error('Cannot power on ' + this.name);
    }

    async setPowerStateOff() {
        this.log('Turning ' + this.name + ' off');
        await this.remote.send('POWER');
    }

    //
    // Speaker service
    //

    get speaker_service() {
        const speaker_service = new this.constructor.hap.Service.Speaker(this.name);

        speaker_service.getCharacteristic(this.constructor.hap.Characteristic.Mute)
            .on('get', callback => this.getMute().then(data => callback(undefined, data)).catch(callback))
            .on('set', (mute, callback) => this.setMute(mute).then(data => callback(undefined, data)).catch(callback));

        speaker_service.getCharacteristic(this.constructor.hap.Characteristic.Volume)
            .on('get', callback => this.getVolume().then(data => callback(undefined, data)).catch(callback))
            .on('set', (volume, callback) => this.setVolume(volume).then(data => callback(undefined, data)).catch(callback));

        return Object.defineProperty(this, 'speaker_service', {configurable: true, value: speaker_service}).speaker_service;
    }

    get speaker_lightbulb_service() {
        const speaker_service = new this.constructor.hap.Service.Lightbulb(this.name + ' Speaker');

        speaker_service.getCharacteristic(this.constructor.hap.Characteristic.On)
            .on('get', callback => this.getMute().then(data => callback(undefined, !data)).catch(callback))
            .on('set', (mute, callback) => this.setMute(!mute).then(data => callback(undefined, data)).catch(callback));

        speaker_service.getCharacteristic(this.constructor.hap.Characteristic.Brightness)
            .on('get', callback => this.getVolume().then(data => callback(undefined, data)).catch(callback))
            .on('set', (volume, callback) => this.setVolume(volume).then(data => callback(undefined, data)).catch(callback));

        return Object.defineProperty(this, 'speaker_lightbulb_service', {configurable: true, value: speaker_service}).speaker_lightbulb_service;
    }

    callMediaRendererAction(service, action, args) {
        if (!this.media_renderer_client) {
            throw new Error('MediaRenderer client not initialized');
        }

        this.log('Calling MediaRenderer action', service, action, args);

        // If the TV doesn't respond in 2 seconds assume the TV is off
        return Promise.race([
            new Promise((resolve, reject) => this.media_renderer_client
                .callAction(service, action, args, (err, data) => err ? reject(err) : resolve(data))),
            new Promise((rs, reject) => setTimeout(reject, 2000)),
        ]);
    }

    handleMediaRendererEvent(event) {
        if (event.Volume) {
            if (this.expose_television_speaker_service) this.tv_speaker_service.getCharacteristic(this.constructor.hap.Characteristic.Volume).updateValue(parseInt(event.Volume));
            if (this.expose_speaker_service) this.speaker_service.getCharacteristic(this.constructor.hap.Characteristic.Volume).updateValue(parseInt(event.Volume));
            if (this.expose_speaker_lightbulb_service) this.speaker_lightbulb_service.getCharacteristic(this.constructor.hap.Characteristic.Brightness).updateValue(parseInt(event.Volume));
        }

        if (event.Mute) {
            if (this.expose_television_speaker_service) this.tv_speaker_service.getCharacteristic(this.constructor.hap.Characteristic.Mute).updateValue(!!parseInt(event.Mute));
            if (this.expose_speaker_service) this.speaker_service.getCharacteristic(this.constructor.hap.Characteristic.Mute).updateValue(!!parseInt(event.Mute));
            if (this.expose_speaker_lightbulb_service) this.speaker_lightbulb_service.getCharacteristic(this.constructor.hap.Characteristic.On).updateValue(!parseInt(event.Mute));
        }
    }

    async getMute() {
        const result = await this.callMediaRendererAction('RenderingControl', 'GetMute', {
            InstanceID: 0,
            Channel: 'Master',
        });

        return !!parseInt(result.CurrentMute);
    }

    async setMute(mute) {
        await this.callMediaRendererAction('RenderingControl', 'SetMute', {
            InstanceID: 0,
            Channel: 'Master',
            DesiredMute: mute,
        });
    }

    async getVolume() {
        const result = await this.callMediaRendererAction('RenderingControl', 'GetVolume', {
            InstanceID: 0,
            Channel: 'Master',
        });

        return parseInt(result.CurrentVolume);
    }

    async setVolume(volume) {
        await this.callMediaRendererAction('RenderingControl', 'SetVolume', {
            InstanceID: 0,
            Channel: 'Master',
            DesiredVolume: volume,
        });
    }

    //
    // Television service
    //

    get tv_service() {
        const tv_service = new this.constructor.hap.Service.Television(this.name);

        tv_service.getCharacteristic(this.constructor.hap.Characteristic.Active)
            .on('get', callback => this.getPowerState().then(data => callback(undefined, data ? this.constructor.hap.Characteristic.Active.ACTIVE : this.constructor.hap.Characteristic.Active.INACTIVE)).catch(callback))
            .on('set', (on, callback) => this.setPowerState(on === this.constructor.hap.Characteristic.Active.ACTIVE ? true : false).then(data => callback(undefined, data)).catch(callback));

        tv_service.getCharacteristic(this.constructor.hap.Characteristic.ActiveIdentifier)
            .on('get', callback => this.getInputIdentifier().then(data => callback(undefined, data)).catch(callback))
            .on('set', (input, callback) => this.setInputIdentifier(input).then(data => callback(undefined, data)).catch(callback));

        tv_service.getCharacteristic(this.constructor.hap.Characteristic.ConfiguredName)
            .on('get', callback => this.getConfiguredName().then(data => callback(undefined, data)).catch(callback))
            .on('set', (name, callback) => this.setConfiguredName(name).then(data => callback(undefined, data)).catch(callback));

        tv_service.setCharacteristic(this.constructor.hap.Characteristic.SleepDiscoveryMode, this.apple_tv ? this.constructor.hap.Characteristic.SleepDiscoveryMode.ALWAYS_DISCOVERABLE : this.constructor.hap.Characteristic.SleepDiscoveryMode.NOT_DISCOVERABLE);

        tv_service.getCharacteristic(this.constructor.hap.Characteristic.RemoteKey)
            .on('set', (key, callback) => this.sendRemoteKey(key).then(data => callback(undefined, data)).catch(callback));

        return Object.defineProperty(this, 'tv_service', {configurable: true, value: tv_service}).tv_service;
    }

    async getInputIdentifier() {
        return this.last_input_switched || this.default_input;
    }

    async setInputIdentifier(identifier) {
        if (!this.input_source_services[identifier]) {
            this.log('Unknown input identifier ' + identifier + ' for ' + this.name);
            return;
        }

        const source = this.tv_sources[identifier];
        const service = this.input_source_services[identifier];

        this.log('Setting input identifier for ' + this.name + ' to ' + identifier + ' (' +
            (await this.getInputSourceName(identifier) || service.displayName) + ')');

        if (source || service.config.channel_number || service.config.send_exit) {
            // Send an EXIT keypress to close the browser if it's open
            await this.remote.send('EXIT');
            await new Promise(r => setTimeout(r, 1000));
        }

        if (source) {
            await this.followtv.setSource(source.source_id);
        } else if (service.config.channel_number) {
            // Switch to the TV input and show this channel
            await this.followtv.setChannel(service.config.channel_number);
        } else if (service.config.remote_key) {
            // Send a keypress to the TV
            await this.remote.send(service.config.remote_key);
        } else if (service.config.app_url) {
            await this.followtv.setURL(service.config.app_url);
        } else {
            this.log('Unsupported source');
            throw new Error('Unsupported source');
        }

        this.last_input_switched = identifier;

        if (service.config.picture_zoom) {
            await new Promise(r => setTimeout(r, 1000));
            await this.followtv.setPictureZoom(service.config.picture_zoom);
        }
    }

    async getConfiguredName() {
        return await this.getItem('ConfiguredName');
    }

    async setConfiguredName(name) {
        this.log('Setting configured name for Television service for ' + this.name + ': ' + name);
        await this.setItem('ConfiguredName', name);
    }

    async sendRemoteKey(key) {
        if (!key_map[key]) {
            throw new Error('Unsupported remote key');
        }

        const name = Object.keys(this.constructor.hap.Characteristic.RemoteKey)[Object.values(this.constructor.hap.Characteristic.RemoteKey).findIndex(value => value === key)];

        this.log('Sending key ' + key_map[key] + ' (' + name + ': ' + key + ') to ' + this.name);

        await this.remote.send(key_map[key]);
    }

    //
    // Input Source services
    //

    get input_source_services() {
        return Object.defineProperty(this, 'input_source_services', {configurable: true, value: []}).input_source_services;
    }

    addInputSourceService(id, config, type, subtype) {
        const input_service = new this.constructor.hap.Service.InputSource(config.name, config.name || id);

        if (!type) {
            if (typeof config === 'object') type = config.input_source_type;
            if (config instanceof Array) type = config[0];
        }
        if (!subtype) {
            if (typeof config === 'object') subtype = config.input_device_type;
            if (config instanceof Array) subtype = config[1];
        }

        if (typeof config !== 'object') config = {};
        input_service.config = config;

        input_service.setCharacteristic(this.constructor.hap.Characteristic.InputSourceType, typeof type === 'number' ? type : this.constructor.hap.Characteristic.InputSourceType[type]);

        if (subtype) {
            input_service.setCharacteristic(this.constructor.hap.Characteristic.InputDeviceType, typeof subtype === 'number' ? subtype : this.constructor.hap.Characteristic.InputDeviceType[subtype]);
        }

        input_service.getCharacteristic(this.constructor.hap.Characteristic.ConfiguredName)
            .on('get', callback => this.getInputSourceName(id).then(data => callback(undefined, data)).catch(callback))
            .on('set', (name, callback) => this.setInputSourceName(id, name).then(data => callback(undefined, data)).catch(callback));

        input_service.setCharacteristic(this.constructor.hap.Characteristic.IsConfigured, this.constructor.hap.Characteristic.IsConfigured.CONFIGURED);
        input_service.setCharacteristic(this.constructor.hap.Characteristic.CurrentVisibilityState, this.constructor.hap.Characteristic.CurrentVisibilityState.SHOWN);

        input_service.setCharacteristic(this.constructor.hap.Characteristic.Identifier, this.input_source_services.length);

        // Remove once iOS 12.2 is released
        input_service.getCharacteristic(this.constructor.hap.Characteristic.On)
            .on('get', callback => callback(undefined, false))
            .on('set', (on, callback) => {
                callback();
                if (!on) return;

                this.tv_service.setCharacteristic(this.constructor.hap.Characteristic.ActiveIdentifier, input_service.getCharacteristic(this.constructor.hap.Characteristic.Identifier).value);
                setTimeout(() => input_service.getCharacteristic(this.constructor.hap.Characteristic.On).updateValue(false), 100);
            });

        this.input_source_services.push(input_service);
        this.tv_service.addLinkedService(input_service);

        return input_service;
    }

    addAppInputSourceService(id, config) {
        if (config.input_source_type === undefined) config.input_source_type = this.constructor.hap.Characteristic.InputSourceType.APPLICATION;

        return this.addInputSourceService(id, config);
    }

    async getInputSourceName(id) {
        return await this.getItem('InputSource.' + id + '.ConfiguredName');
    }

    async setInputSourceName(id, name) {
        this.log('Setting configured name for Input Source service "' + id + '" for ' + this.name + ': ' + name);
        await this.setItem('InputSource.' + id + '.ConfiguredName', name);
    }

    //
    // Television Speaker service
    //

    get tv_speaker_service() {
        const speaker_service = new this.constructor.hap.Service.TelevisionSpeaker(this.name);

        speaker_service.getCharacteristic(this.constructor.hap.Characteristic.Mute)
            .on('get', callback => this.getMute().then(data => callback(undefined, data)).catch(callback))
            .on('set', (mute, callback) => this.setMute(mute).then(data => callback(undefined, data)).catch(callback));

        speaker_service.getCharacteristic(this.constructor.hap.Characteristic.Volume)
            .on('get', callback => this.getVolume().then(data => callback(undefined, data)).catch(callback))
            .on('set', (volume, callback) => this.setVolume(volume).then(data => callback(undefined, data)).catch(callback));

        speaker_service.setCharacteristic(this.constructor.hap.Characteristic.VolumeControlType, this.constructor.hap.Characteristic.VolumeControlType.ABSOLUTE);

        return Object.defineProperty(this, 'tv_speaker_service', {configurable: true, value: speaker_service}).tv_speaker_service;
    }
}

module.exports.TVAccessory = TVAccessory;
module.exports.key_map = key_map;
